import json
import time
import asyncio
from datetime import datetime
from typing import Dict, Any, Optional, List

import urllib3
import aiosonic

from kakaowork.consts import (
    Limit,
    BASE_URL,
    BASE_PATH_USERS,
    BASE_PATH_CONVERSATIONS,
    BASE_PATH_MESSAGES,
    BASE_PATH_DEPARTMENTS,
    BASE_PATH_SPACES,
    BASE_PATH_BOTS,
    BASE_PATH_BATCH,
)
from kakaowork.models import (
    WorkTimeField,
    VacationTimeField,
    BaseResponse,
    UserResponse,
    UserListResponse,
    ConversationResponse,
    ConversationListResponse,
    MessageResponse,
    DepartmentListResponse,
    SpaceResponse,
    BotResponse,
)
from kakaowork.blockkit import Block
from kakaowork.utils import json_default, drop_none
from kakaowork.ratelimit import RateLimiter


class Kakaowork:
    class Users:
        def __init__(self, client: 'Kakaowork', *, base_path: Optional[str] = BASE_PATH_USERS):
            self.client = client
            self.base_path = base_path

        def info(self, *, user_id: int) -> UserResponse:
            with self.client.limiter:
                r = self.client.http.request(
                    'GET',
                    f'{self.client.base_url}{self.base_path}.info',
                    fields={'user_id': user_id},
                )
            self.client._respect_rate_limit(r)
            return UserResponse.parse_raw(r.data)

        def find_by_email(self, email: str) -> UserResponse:
            with self.client.limiter:
                r = self.client.http.request(
                    'GET',
                    f'{self.client.base_url}{self.base_path}.find_by_email',
                    fields={'email': email},
                )
            self.client._respect_rate_limit(r)
            return UserResponse.parse_raw(r.data)

        def find_by_phone_number(self, phone_number: str) -> UserResponse:
            with self.client.limiter:
                r = self.client.http.request(
                    'GET',
                    f'{self.client.base_url}{self.base_path}.find_by_phone_number',
                    fields={'phone_number': phone_number},
                )
            self.client._respect_rate_limit(r)
            return UserResponse.parse_raw(r.data)

        def list(self, *, cursor: Optional[str] = None, limit: Optional[int] = Limit.DEFAULT) -> UserListResponse:
            fields: Dict[str, Any] = {'cursor': cursor} if cursor else {'limit': str(limit)}
            with self.client.limiter:
                r = self.client.http.request(
                    'GET',
                    f'{self.client.base_url}{self.base_path}.list',
                    fields=fields,
                )
            self.client._respect_rate_limit(r)
            return UserListResponse.parse_raw(r.data)

        def set_work_time(self, *, user_id: int, work_start_time: datetime, work_end_time: datetime) -> BaseResponse:
            payload = {
                'user_id': user_id,
                'work_start_time': int(work_start_time.timestamp()),
                'work_end_time': int(work_end_time.timestamp()),
            }
            with self.client.limiter:
                r = self.client.http.request(
                    'POST',
                    f'{self.client.base_url}{self.base_path}.set_work_time',
                    body=json.dumps(payload).encode('utf-8'),
                )
            self.client._respect_rate_limit(r)
            return BaseResponse.parse_raw(r.data)

        def set_vacation_time(self, *, user_id: int, vacation_start_time: datetime, vacation_end_time: datetime) -> BaseResponse:
            payload = {
                'user_id': user_id,
                'vacation_start_time': int(vacation_start_time.timestamp()),
                'vacation_end_time': int(vacation_end_time.timestamp()),
            }
            with self.client.limiter:
                r = self.client.http.request(
                    'POST',
                    f'{self.client.base_url}{self.base_path}.set_vacation_time',
                    body=json.dumps(payload).encode('utf-8'),
                )
            self.client._respect_rate_limit(r)
            return BaseResponse.parse_raw(r.data)

    class Conversations:
        def __init__(self, client: 'Kakaowork', *, base_path: Optional[str] = BASE_PATH_CONVERSATIONS):
            self.client = client
            self.base_path = base_path

        def open(self, *, user_ids: List[int]) -> ConversationResponse:
            payload: Dict[str, Any] = {'user_id': user_ids[0]} if len(user_ids) == 1 else {'user_ids': user_ids}
            with self.client.limiter:
                r = self.client.http.request(
                    'POST',
                    f'{self.client.base_url}{self.base_path}.open',
                    body=json.dumps(payload).encode('utf-8'),
                )
            self.client._respect_rate_limit(r)
            return ConversationResponse.parse_raw(r.data)

        def list(self, *, cursor: Optional[str] = None, limit: Optional[int] = Limit.DEFAULT) -> ConversationListResponse:
            fields = {'cursor': cursor} if cursor else {'limit': str(limit)}
            with self.client.limiter:
                r = self.client.http.request(
                    'GET',
                    f'{self.client.base_url}{self.base_path}.list',
                    fields=fields,
                )
            self.client._respect_rate_limit(r)
            return ConversationListResponse.parse_raw(r.data)

        def users(self, *, conversation_id: int) -> UserListResponse:
            with self.client.limiter:
                r = self.client.http.request(
                    'GET',
                    f'{self.client.base_url}{self.base_path}/{conversation_id}/users',
                )
            self.client._respect_rate_limit(r)
            return UserListResponse.parse_raw(r.data)

        def invite(self, *, conversation_id: int, user_ids: List[int]) -> BaseResponse:
            payload = {'user_ids': user_ids}
            with self.client.limiter:
                r = self.client.http.request(
                    'POST',
                    f'{self.client.base_url}{self.base_path}/{conversation_id}/invite',
                    body=json.dumps(payload).encode('utf-8'),
                )
            self.client._respect_rate_limit(r)
            return BaseResponse.parse_raw(r.data)

        def kick(self, *, conversation_id: int, user_ids: List[int]) -> BaseResponse:
            payload = {'user_ids': user_ids}
            with self.client.limiter:
                r = self.client.http.request(
                    'POST',
                    f'{self.client.base_url}{self.base_path}/{conversation_id}/kick',
                    body=json.dumps(payload).encode('utf-8'),
                )
            self.client._respect_rate_limit(r)
            return BaseResponse.parse_raw(r.data)

    class Messages:
        def __init__(self, client: 'Kakaowork', *, base_path: Optional[str] = BASE_PATH_MESSAGES):
            self.client = client
            self.base_path = base_path

        def send(self, *, conversation_id: int, text: str, blocks: Optional[List[Block]] = None) -> MessageResponse:
            payload = drop_none({
                'conversation_id': conversation_id,
                'text': text,
                'blocks': blocks,
            })
            with self.client.limiter:
                r = self.client.http.request(
                    'POST',
                    f'{self.client.base_url}{self.base_path}.send',
                    body=json.dumps(payload, default=json_default).encode('utf-8'),
                )
            self.client._respect_rate_limit(r)
            return MessageResponse.parse_raw(r.data)

        def send_by(self, *, text: str, email: Optional[str] = None, key: Optional[str] = None, blocks: Optional[List[Block]] = None) -> MessageResponse:
            if not (email or key):
                raise ValueError("Either 'email' or 'key' must exist.")
            payload = drop_none({
                'email': email,
                'key': key,
                'text': text,
                'blocks': blocks,
            })
            with self.client.limiter:
                r = self.client.http.request(
                    'POST',
                    f'{self.client.base_url}{self.base_path}.send_by',
                    body=json.dumps(payload, default=json_default).encode('utf-8'),
                )
            self.client._respect_rate_limit(r)
            return MessageResponse.parse_raw(r.data)

        def send_by_email(self, email: str, *, text: str, blocks: Optional[List[Block]] = None) -> MessageResponse:
            payload = drop_none({
                'email': email,
                'text': text,
                'blocks': blocks,
            })
            with self.client.limiter:
                r = self.client.http.request(
                    'POST',
                    f'{self.client.base_url}{self.base_path}.send_by_email',
                    body=json.dumps(payload, default=json_default).encode('utf-8'),
                )
            self.client._respect_rate_limit(r)
            return MessageResponse.parse_raw(r.data)

    class Departments:
        def __init__(self, client: 'Kakaowork', *, base_path: Optional[str] = BASE_PATH_DEPARTMENTS):
            self.client = client
            self.base_path = base_path

        def list(self, *, cursor: Optional[str] = None, limit: Optional[int] = Limit.DEFAULT) -> DepartmentListResponse:
            fields = {'cursor': cursor} if cursor else {'limit': str(limit)}
            with self.client.limiter:
                r = self.client.http.request(
                    'GET',
                    f'{self.client.base_url}{self.base_path}.list',
                    fields=fields,
                )
            self.client._respect_rate_limit(r)
            return DepartmentListResponse.parse_raw(r.data)

    class Spaces:
        def __init__(self, client: 'Kakaowork', *, base_path: Optional[str] = BASE_PATH_SPACES):
            self.client = client
            self.base_path = base_path

        def info(self) -> SpaceResponse:
            with self.client.limiter:
                r = self.client.http.request(
                    'GET',
                    f'{self.client.base_url}{self.base_path}.info',
                )
            self.client._respect_rate_limit(r)
            return SpaceResponse.parse_raw(r.data)

    class Bots:
        def __init__(self, client: 'Kakaowork', *, base_path: Optional[str] = BASE_PATH_BOTS):
            self.client = client
            self.base_path = base_path

        def info(self) -> BotResponse:
            with self.client.limiter:
                r = self.client.http.request(
                    'GET',
                    f'{self.client.base_url}{self.base_path}.info',
                )
            self.client._respect_rate_limit(r)
            return BotResponse.parse_raw(r.data)

    def __init__(self, *, app_key: str, base_url: Optional[str] = BASE_URL):
        self.app_key = app_key
        self.base_url = base_url
        self.http = urllib3.PoolManager(headers=self.headers, retries=3, maxsize=5)
        self.limiter = RateLimiter(capacity=0, refill_rate=60.0)

    def _respect_rate_limit(self, response: urllib3.HTTPResponse) -> None:
        if 200 <= response.status < 300 and self.limiter.capacity <= 0:
            capacity = int(response.headers.get('ratelimit-limit', 0))
            self.limiter.reset(capacity=capacity)
        elif response.status == 429:
            capacity = int(response.headers.get('ratelimit-limit', 0))
            wait_time = int(response.headers.get('retry-after', 0))
            time.sleep(wait_time)
            self.limiter.reset(capacity=capacity)

    class Batch:
        def __init__(self, client: 'Kakaowork', *, base_path: Optional[str] = BASE_PATH_BATCH):
            self.client = client
            self.base_path = base_path

        class Users:
            def __init__(self, batch: 'Kakaowork.Batch'):
                self.client = batch.client
                self.base_path = f'{batch.base_path}/users'

            def set_work_time(self, items: List[WorkTimeField]) -> BaseResponse:
                payload = {
                    'user_work_times': [item.dict(exclude_none=True) for item in items],
                }
                with self.client.limiter:
                    r = self.client.http.request(
                        'POST',
                        f'{self.client.base_url}{self.base_path}.set_work_time',
                        body=json.dumps(payload, default=json_default).encode('utf-8'),
                    )
                self.client._respect_rate_limit(r)
                return BaseResponse.parse_raw(r.data)

            def set_vacation_time(self, items: List[VacationTimeField]) -> BaseResponse:
                payload = {
                    'user_vacation_times': [item.dict(exclude_none=True) for item in items],
                }
                with self.client.limiter:
                    r = self.client.http.request(
                        'POST',
                        f'{self.client.base_url}{self.base_path}.set_vacation_time',
                        body=json.dumps(payload, default=json_default).encode('utf-8'),
                    )
                self.client._respect_rate_limit(r)
                return BaseResponse.parse_raw(r.data)

            def reset_work_time(self, *, user_ids: List[int]) -> BaseResponse:
                payload = {
                    'user_ids': user_ids,
                }
                with self.client.limiter:
                    r = self.client.http.request(
                        'POST',
                        f'{self.client.base_url}{self.base_path}.reset_work_time',
                        body=json.dumps(payload).encode('utf-8'),
                    )
                self.client._respect_rate_limit(r)
                return BaseResponse.parse_raw(r.data)

            def reset_vacation_time(self, *, user_ids: List[int]) -> BaseResponse:
                payload = {
                    'user_ids': user_ids,
                }
                with self.client.limiter:
                    r = self.client.http.request(
                        'POST',
                        f'{self.client.base_url}{self.base_path}.reset_vacation_time',
                        body=json.dumps(payload).encode('utf-8'),
                    )
                self.client._respect_rate_limit(r)
                return BaseResponse.parse_raw(r.data)

        @property
        def users(self) -> Users:
            return self.Users(self)

    @property
    def headers(self) -> Dict[str, Any]:
        return {
            'Authorization': f'Bearer {self.app_key}',
            'Content-Type': 'application/json; charset=utf-8',
        }

    @property
    def users(self) -> Users:
        return self.Users(self)

    @property
    def conversations(self) -> Conversations:
        return self.Conversations(self)

    @property
    def messages(self) -> Messages:
        return self.Messages(self)

    @property
    def departments(self) -> Departments:
        return self.Departments(self)

    @property
    def spaces(self) -> Spaces:
        return self.Spaces(self)

    @property
    def bots(self) -> Bots:
        return self.Bots(self)

    @property
    def batch(self) -> Batch:
        return self.Batch(self)


class AsyncKakaowork:
    class Users:
        def __init__(self, client: 'AsyncKakaowork', *, base_path: Optional[str] = BASE_PATH_USERS):
            self.client = client
            self.base_path = base_path

        async def info(self, *, user_id: int) -> UserResponse:
            async with self.client.limiter:
                r = await self.client.http.request(
                    url=f'{self.client.base_url}{self.base_path}.info',
                    method='GET',
                    headers=self.client.headers,
                    params={'user_id': user_id},
                )
            await self.client._respect_rate_limit(r)
            return UserResponse.parse_raw(await r.content())

        async def find_by_email(self, email: str) -> UserResponse:
            async with self.client.limiter:
                r = await self.client.http.request(
                    url=f'{self.client.base_url}{self.base_path}.find_by_email',
                    method='GET',
                    headers=self.client.headers,
                    params={'email': email},
                )
            await self.client._respect_rate_limit(r)
            return UserResponse.parse_raw(await r.content())

        async def find_by_phone_number(self, phone_number: str) -> UserResponse:
            async with self.client.limiter:
                r = await self.client.http.request(
                    url=f'{self.client.base_url}{self.base_path}.find_by_phone_number',
                    method='GET',
                    headers=self.client.headers,
                    params={'phone_number': phone_number},
                )
            await self.client._respect_rate_limit(r)
            return UserResponse.parse_raw(await r.content())

        async def list(self, *, cursor: Optional[str] = None, limit: Optional[int] = Limit.DEFAULT) -> UserListResponse:
            params: Dict[str, Any] = {'cursor': cursor} if cursor else {'limit': str(limit)}
            async with self.client.limiter:
                r = await self.client.http.request(
                    url=f'{self.client.base_url}{self.base_path}.list',
                    method='GET',
                    headers=self.client.headers,
                    params=params,
                )
            await self.client._respect_rate_limit(r)
            return UserListResponse.parse_raw(await r.content())

        async def set_work_time(self, *, user_id: int, work_start_time: datetime, work_end_time: datetime) -> BaseResponse:
            payload = {
                'user_id': user_id,
                'work_start_time': int(work_start_time.timestamp()),
                'work_end_time': int(work_end_time.timestamp()),
            }
            async with self.client.limiter:
                r = await self.client.http.request(
                    url=f'{self.client.base_url}{self.base_path}.set_work_time',
                    method='POST',
                    headers=self.client.headers,
                    data=json.dumps(payload).encode('utf-8'),
                )
            await self.client._respect_rate_limit(r)
            return BaseResponse.parse_raw(await r.content())

        async def set_vacation_time(self, *, user_id: int, vacation_start_time: datetime, vacation_end_time: datetime) -> BaseResponse:
            payload = {
                'user_id': user_id,
                'vacation_start_time': int(vacation_start_time.timestamp()),
                'vacation_end_time': int(vacation_end_time.timestamp()),
            }
            async with self.client.limiter:
                r = await self.client.http.request(
                    url=f'{self.client.base_url}{self.base_path}.set_vacation_time',
                    method='POST',
                    headers=self.client.headers,
                    data=json.dumps(payload).encode('utf-8'),
                )
            await self.client._respect_rate_limit(r)
            return BaseResponse.parse_raw(await r.content())

    class Conversations:
        def __init__(self, client: 'AsyncKakaowork', *, base_path: Optional[str] = BASE_PATH_CONVERSATIONS):
            self.client = client
            self.base_path = base_path

        async def open(self, *, user_ids: List[int]) -> ConversationResponse:
            payload: Dict[str, Any] = {'user_id': user_ids[0]} if len(user_ids) == 1 else {'user_ids': user_ids}
            async with self.client.limiter:
                r = await self.client.http.request(
                    url=f'{self.client.base_url}{self.base_path}.open',
                    method='POST',
                    headers=self.client.headers,
                    data=json.dumps(payload).encode('utf-8'),
                )
            await self.client._respect_rate_limit(r)
            return ConversationResponse.parse_raw(await r.content())

        async def list(self, *, cursor: Optional[str] = None, limit: Optional[int] = Limit.DEFAULT) -> ConversationListResponse:
            params = {'cursor': cursor} if cursor else {'limit': str(limit)}
            async with self.client.limiter:
                r = await self.client.http.request(
                    url=f'{self.client.base_url}{self.base_path}.list',
                    method='GET',
                    headers=self.client.headers,
                    params=params,
                )
            await self.client._respect_rate_limit(r)
            return ConversationListResponse.parse_raw(await r.content())

        async def users(self, *, conversation_id: int) -> UserListResponse:
            async with self.client.limiter:
                r = await self.client.http.request(
                    url=f'{self.client.base_url}{self.base_path}/{conversation_id}/users',
                    method='GET',
                    headers=self.client.headers,
                )
            await self.client._respect_rate_limit(r)
            return UserListResponse.parse_raw(await r.content())

        async def invite(self, *, conversation_id: int, user_ids: List[int]) -> BaseResponse:
            payload = {'user_ids': user_ids}
            async with self.client.limiter:
                r = await self.client.http.request(
                    url=f'{self.client.base_url}{self.base_path}/{conversation_id}/invite',
                    method='POST',
                    headers=self.client.headers,
                    data=json.dumps(payload).encode('utf-8'),
                )
            await self.client._respect_rate_limit(r)
            return BaseResponse.parse_raw(await r.content())

        async def kick(self, *, conversation_id: int, user_ids: List[int]) -> BaseResponse:
            payload = {'user_ids': user_ids}
            async with self.client.limiter:
                r = await self.client.http.request(
                    url=f'{self.client.base_url}{self.base_path}/{conversation_id}/kick',
                    method='POST',
                    headers=self.client.headers,
                    data=json.dumps(payload).encode('utf-8'),
                )
            await self.client._respect_rate_limit(r)
            return BaseResponse.parse_raw(await r.content())

    class Messages:
        def __init__(self, client: 'AsyncKakaowork', *, base_path: Optional[str] = BASE_PATH_MESSAGES):
            self.client = client
            self.base_path = base_path

        async def send(self, *, conversation_id: int, text: str, blocks: Optional[List[Block]] = None) -> MessageResponse:
            payload = drop_none({
                'conversation_id': conversation_id,
                'text': text,
                'blocks': blocks,
            })
            async with self.client.limiter:
                r = await self.client.http.request(
                    url=f'{self.client.base_url}{self.base_path}.send',
                    method='POST',
                    headers=self.client.headers,
                    data=json.dumps(payload, default=json_default).encode('utf-8'),
                )
            await self.client._respect_rate_limit(r)
            return MessageResponse.parse_raw(await r.content())

        async def send_by(self, *, text: str, email: Optional[str] = None, key: Optional[str] = None, blocks: Optional[List[Block]] = None) -> MessageResponse:
            if not (email or key):
                raise ValueError("Either 'email' or 'key' must exist.")
            payload = drop_none({
                'email': email,
                'key': key,
                'text': text,
                'blocks': blocks,
            })
            async with self.client.limiter:
                r = await self.client.http.request(
                    url=f'{self.client.base_url}{self.base_path}.send_by',
                    method='POST',
                    headers=self.client.headers,
                    data=json.dumps(payload, default=json_default).encode('utf-8'),
                )
            await self.client._respect_rate_limit(r)
            return MessageResponse.parse_raw(await r.content())

        async def send_by_email(self, email: str, *, text: str, blocks: Optional[List[Block]] = None) -> MessageResponse:
            payload = drop_none({
                'email': email,
                'text': text,
                'blocks': blocks,
            })
            async with self.client.limiter:
                r = await self.client.http.request(
                    url=f'{self.client.base_url}{self.base_path}.send_by_email',
                    method='POST',
                    headers=self.client.headers,
                    data=json.dumps(payload, default=json_default).encode('utf-8'),
                )
            await self.client._respect_rate_limit(r)
            return MessageResponse.parse_raw(await r.content())

    class Departments:
        def __init__(self, client: 'AsyncKakaowork', *, base_path: Optional[str] = BASE_PATH_DEPARTMENTS):
            self.client = client
            self.base_path = base_path

        async def list(self, *, cursor: Optional[str] = None, limit: Optional[int] = Limit.DEFAULT) -> DepartmentListResponse:
            params = {'cursor': cursor} if cursor else {'limit': str(limit)}
            async with self.client.limiter:
                r = await self.client.http.request(
                    url=f'{self.client.base_url}{self.base_path}.list',
                    method='GET',
                    headers=self.client.headers,
                    params=params,
                )
            await self.client._respect_rate_limit(r)
            return DepartmentListResponse.parse_raw(await r.content())

    class Spaces:
        def __init__(self, client: 'AsyncKakaowork', *, base_path: Optional[str] = BASE_PATH_SPACES):
            self.client = client
            self.base_path = base_path

        async def info(self) -> SpaceResponse:
            async with self.client.limiter:
                r = await self.client.http.request(
                    url=f'{self.client.base_url}{self.base_path}.info',
                    method='GET',
                    headers=self.client.headers,
                )
            await self.client._respect_rate_limit(r)
            return SpaceResponse.parse_raw(await r.content())

    class Bots:
        def __init__(self, client: 'AsyncKakaowork', *, base_path: Optional[str] = BASE_PATH_BOTS):
            self.client = client
            self.base_path = base_path

        async def info(self) -> BotResponse:
            async with self.client.limiter:
                r = await self.client.http.request(
                    url=f'{self.client.base_url}{self.base_path}.info',
                    method='GET',
                    headers=self.client.headers,
                )
            await self.client._respect_rate_limit(r)
            return BotResponse.parse_raw(await r.content())

    class Batch:
        def __init__(self, client: 'AsyncKakaowork', *, base_path: Optional[str] = BASE_PATH_BATCH):
            self.client = client
            self.base_path = base_path

        class Users:
            def __init__(self, batch: 'AsyncKakaowork.Batch'):
                self.client = batch.client
                self.base_path = f'{batch.base_path}/users'

            async def set_work_time(self, items: List[WorkTimeField]) -> BaseResponse:
                payload = {
                    'user_work_times': [item.dict(exclude_none=True) for item in items],
                }
                async with self.client.limiter:
                    r = await self.client.http.request(
                        url=f'{self.client.base_url}{self.base_path}.set_work_time',
                        method='POST',
                        headers=self.client.headers,
                        data=json.dumps(payload, default=json_default).encode('utf-8'),
                    )
                await self.client._respect_rate_limit(r)
                return BaseResponse.parse_raw(await r.content())

            async def set_vacation_time(self, items: List[VacationTimeField]) -> BaseResponse:
                payload = {
                    'user_vacation_times': [item.dict(exclude_none=True) for item in items],
                }
                async with self.client.limiter:
                    r = await self.client.http.request(
                        url=f'{self.client.base_url}{self.base_path}.set_vacation_time',
                        method='POST',
                        headers=self.client.headers,
                        data=json.dumps(payload, default=json_default).encode('utf-8'),
                    )
                await self.client._respect_rate_limit(r)
                return BaseResponse.parse_raw(await r.content())

            async def reset_work_time(self, *, user_ids: List[int]) -> BaseResponse:
                payload = {
                    'user_ids': user_ids,
                }
                async with self.client.limiter:
                    r = await self.client.http.request(
                        url=f'{self.client.base_url}{self.base_path}.reset_work_time',
                        method='POST',
                        headers=self.client.headers,
                        data=json.dumps(payload).encode('utf-8'),
                    )
                await self.client._respect_rate_limit(r)
                return BaseResponse.parse_raw(await r.content())

            async def reset_vacation_time(self, *, user_ids: List[int]) -> BaseResponse:
                payload = {
                    'user_ids': user_ids,
                }
                async with self.client.limiter:
                    r = await self.client.http.request(
                        url=f'{self.client.base_url}{self.base_path}.reset_vacation_time',
                        method='POST',
                        headers=self.client.headers,
                        data=json.dumps(payload).encode('utf-8'),
                    )
                await self.client._respect_rate_limit(r)
                return BaseResponse.parse_raw(await r.content())

        @property
        def users(self) -> Users:
            return self.Users(self)

    def __init__(self, *, app_key: str, base_url: Optional[str] = BASE_URL):
        self.app_key = app_key
        self.base_url = base_url
        self.http = aiosonic.HTTPClient()
        self.limiter = RateLimiter(capacity=0, refill_rate=60.0)

    async def _respect_rate_limit(self, response: aiosonic.HttpResponse) -> None:
        if 200 <= response.status_code < 300 and self.limiter.capacity <= 0:
            capacity = int(response.headers.get('ratelimit-limit', 0))
            self.limiter.reset(capacity=capacity)
        elif response.status_code == 429:
            capacity = int(response.headers.get('ratelimit-limit', 0))
            wait_time = int(response.headers.get('retry-after', 0))
            await asyncio.sleep(wait_time)
            self.limiter.reset(capacity=capacity)

    @property
    def headers(self) -> Dict[str, Any]:
        return {
            'Authorization': f'Bearer {self.app_key}',
            'Content-Type': 'application/json; charset=utf-8',
        }

    @property
    def users(self) -> Users:
        return self.Users(self)

    @property
    def conversations(self) -> Conversations:
        return self.Conversations(self)

    @property
    def messages(self) -> Messages:
        return self.Messages(self)

    @property
    def departments(self) -> Departments:
        return self.Departments(self)

    @property
    def spaces(self) -> Spaces:
        return self.Spaces(self)

    @property
    def bots(self) -> Bots:
        return self.Bots(self)

    @property
    def batch(self) -> Batch:
        return self.Batch(self)



from abc import ABCMeta, abstractmethod

from kakaowork.models import SubmitActionReactiveBody, SubmitModalReactiveBody, RequestModalReactiveBody, RequestModalReactiveResponse


class BaseReactiveActionHandler(metaclass=ABCMeta):
    """An abstract class for reactive action handler."""
    @abstractmethod
    def handle_submit(self, body: SubmitActionReactiveBody) -> bool:
        """An abstract method for handling user's request from the KakaoWork server to your server.

        Args:
            body: The request body of user's submission.

        Returns:
            True if this handler succeeds, False otherwise.
        """
        raise NotImplementedError()


class BaseReactiveModalHandler(metaclass=ABCMeta):
    """An abstract class for reactive modal handler."""
    @abstractmethod
    def handle_request(self, body: RequestModalReactiveBody) -> RequestModalReactiveResponse:
        """An abstract method for handling a request to compose a modal from the KakaoWork server to your server.

        Args:
            body: The request body of user's modal action.

        Returns:
            True if this handler succeeds, False otherwise.
        """
        raise NotImplementedError()

    @abstractmethod
    def handle_submit(self, body: SubmitModalReactiveBody) -> bool:
        """An abstract method for handling user's request from the KakaoWork server to your server.

        Args:
            body: The request body of user's submission.

        Returns:
            True if this handler succeeds, False otherwise.
        """
        raise NotImplementedError()



from abc import ABC
from datetime import datetime
from typing import Optional, Union, Any, Dict, List

from pydantic import BaseModel, validator

from kakaowork.consts import StrEnum
from kakaowork.blockkit import Block
from kakaowork.utils import to_kst


class ErrorCode(StrEnum):
    # If you return the 'UNKNOWN' error code, even though it's an official error code. please report it!
    UNKNOWN = 'unknown'

    # Common error codes
    INVALID_PARAMETER = 'invalid_parameter'
    INVALID_AUTHENTICATION = 'invalid_authentication'
    INVALID_REPRESENTATION = 'invalid_representation'
    INVALID_CONTENT_TYPE = 'invalid_content_type'
    API_NOT_FOUND = 'api_not_found'
    UNAUTHORIZED = 'unauthorized'
    INTERNAL_SERVER_ERROR = 'internal_server_error'
    TOO_MANY_REQUESTS = 'too_many_requests'
    EXPIRED_AUTHENTICATION = 'expired_authentication'
    MISSING_PARAMETER = 'missing_parameter'
    BAD_REQUEST = 'bad_request'

    # API specific codes
    USER_NOT_FOUND = 'user_not_found'
    CONVERSATION_NOT_FOUND = 'conversation_not_found'
    TEXT_TOO_LONG = 'text_too_long'
    INVALID_BLOCKS = 'invalid_blocks'

    @classmethod
    def _missing_(cls, value: Any) -> 'ErrorCode':
        return cls.UNKNOWN


class ConversationType(StrEnum):
    DM = 'dm'
    GROUP = 'group'


class ColorTone(StrEnum):
    LIGHT = 'light'
    DARK = 'dark'


class ProfileNameFormat(StrEnum):
    NAME_ONLY = 'name_only'
    NAME_NICKNAME = 'name_nickname'
    NICKNAME_NAME = 'nickname_name'


class ProfilePositionFormat(StrEnum):
    POSITION = 'position'
    RESPONSIBILITY = 'responsibility'


class BotStatus(StrEnum):
    ACTIVATED = 'activated'
    DEACTIVATED = 'deactivated'


class ErrorField(BaseModel):
    code: ErrorCode
    message: str


class UserIdentificationField(BaseModel):
    type: str
    value: str


class UserField(BaseModel):
    id: str
    space_id: str
    name: str
    display_name: Optional[str] = None
    identifications: Optional[List[UserIdentificationField]] = None
    nickname: Optional[str] = None
    avatar_url: Optional[str] = None
    department: Optional[str] = None
    position: Optional[str] = None
    responsibility: Optional[str] = None
    tels: Optional[List[str]] = None
    mobiles: Optional[List[str]] = None
    work_start_time: Optional[datetime] = None
    work_end_time: Optional[datetime] = None
    vacation_start_time: Optional[datetime] = None
    vacation_end_time: Optional[datetime] = None

    class Config:
        json_encoders = {
            datetime: lambda dt: int(dt.timestamp()),
        }

    def __init__(self,
                 *,
                 work_start_time: Optional[Union[datetime, int, str]] = None,
                 work_end_time: Optional[Union[datetime, int, str]] = None,
                 vacation_start_time: Optional[Union[datetime, int, str]] = None,
                 vacation_end_time: Optional[Union[datetime, int, str]] = None,
                 **data) -> None:
        if work_start_time:
            data['work_start_time'] = to_kst(work_start_time) if isinstance(work_start_time, (datetime, int)) else work_start_time
        if work_end_time:
            data['work_end_time'] = to_kst(work_end_time) if isinstance(work_end_time, (datetime, int)) else work_end_time
        if vacation_start_time:
            data['vacation_start_time'] = to_kst(vacation_start_time) if isinstance(vacation_start_time, (datetime, int)) else vacation_start_time
        if vacation_end_time:
            data['vacation_end_time'] = to_kst(vacation_end_time) if isinstance(vacation_end_time, (datetime, int)) else vacation_end_time
        super().__init__(**data)


class ConversationField(BaseModel):
    id: str
    type: ConversationType
    users_count: int
    avatar_url: Optional[str] = None
    name: Optional[str] = None


class MessageField(BaseModel):
    id: str
    text: str
    user_id: str
    conversation_id: int
    send_time: datetime
    update_time: datetime
    blocks: Optional[List[Block]] = None

    class Config:
        json_encoders = {
            datetime: lambda dt: int(dt.timestamp()),
        }

    def __init__(self,
                 *,
                 send_time: Optional[Union[datetime, int, str]] = None,
                 update_time: Optional[Union[datetime, int, str]] = None,
                 blocks: Optional[List[Union[Block, Dict]]] = None,
                 **data) -> None:
        if send_time:
            data['send_time'] = to_kst(send_time) if isinstance(send_time, (datetime, int)) else send_time
        if update_time:
            data['update_time'] = to_kst(update_time) if isinstance(update_time, (datetime, int)) else update_time
        if blocks is not None:
            data['blocks'] = [Block.new(block) for block in blocks]
        super().__init__(**data)


class DepartmentField(BaseModel):
    id: str
    ids_path: str
    parent_id: str
    space_id: str
    name: str
    code: str
    user_count: int
    has_child: Optional[bool] = None
    depth: Optional[int] = None
    users_ids: Optional[List[int]] = None
    leader_ids: Optional[List[int]] = None
    ancestry: Optional[str] = None


class SpaceField(BaseModel):
    id: int
    kakaoi_org_id: int
    name: str
    color_code: str
    color_tone: ColorTone
    permitted_ext: List[str]
    profile_name_format: ProfileNameFormat
    profile_position_format: ProfilePositionFormat
    logo_url: str


class BotField(BaseModel):
    bot_id: int
    title: str
    status: BotStatus


class WorkTimeField(BaseModel):
    user_id: int
    work_start_time: datetime
    work_end_time: datetime

    class Config:
        json_encoders = {
            datetime: lambda dt: int(dt.timestamp()),
        }

    def __init__(
        self,
        *,
        work_start_time: Optional[Union[datetime, int, str]] = None,
        work_end_time: Optional[Union[datetime, int, str]] = None,
        **data,
    ) -> None:
        if work_start_time:
            data['work_start_time'] = to_kst(work_start_time) if isinstance(work_start_time, (datetime, int)) else work_start_time
        if work_end_time:
            data['work_end_time'] = to_kst(work_end_time) if isinstance(work_end_time, (datetime, int)) else work_end_time
        super().__init__(**data)


class VacationTimeField(BaseModel):
    user_id: int
    vacation_start_time: datetime
    vacation_end_time: datetime

    class Config:
        json_encoders = {
            datetime: lambda dt: int(dt.timestamp()),
        }

    def __init__(
        self,
        *,
        vacation_start_time: Optional[Union[datetime, int, str]] = None,
        vacation_end_time: Optional[Union[datetime, int, str]] = None,
        **data,
    ) -> None:
        if vacation_start_time:
            data['vacation_start_time'] = to_kst(vacation_start_time) if isinstance(vacation_start_time, (datetime, int)) else vacation_start_time
        if vacation_end_time:
            data['vacation_end_time'] = to_kst(vacation_end_time) if isinstance(vacation_end_time, (datetime, int)) else vacation_end_time
        super().__init__(**data)


class ReactiveType(StrEnum):
    SUBMIT_ACTION = "submit_action"
    SUBMIT_MODAL = "submission"
    REQUEST_MODAL = "request_modal"


class BaseReactiveBody(BaseModel, ABC):
    type: ReactiveType
    action_time: str
    message: MessageField
    value: str

    class Config:
        validate_assignment = True
        json_encoders = {
            datetime: lambda dt: int(dt.timestamp()),
        }

    def __str__(self):
        return self.json(exclude_none=True)

    def __repr__(self):
        return str(self)

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, BaseReactiveBody):
            return False
        return self.dict(exclude_none=True) == value.dict(exclude_none=True)


class SubmitActionReactiveBody(BaseReactiveBody):
    action_name: str
    react_user_id: int

    def __init__(self, **data) -> None:
        if 'type' not in data:
            data['type'] = ReactiveType.SUBMIT_ACTION
        super().__init__(**data)

    @validator('type')
    def _check_type(cls, value: ReactiveType) -> ReactiveType:
        if value != ReactiveType.SUBMIT_ACTION:
            raise ValueError(f"The 'type' should be f{ReactiveType.SUBMIT_ACTION}")
        return value


class SubmitModalReactiveBody(BaseReactiveBody):
    actions: Dict[str, Any]
    react_user_id: int

    def __init__(self, **data) -> None:
        if 'type' not in data:
            data['type'] = ReactiveType.SUBMIT_MODAL
        super().__init__(**data)

    @validator('type')
    def _check_type(cls, value: ReactiveType) -> ReactiveType:
        if value != ReactiveType.SUBMIT_MODAL:
            raise ValueError(f"The 'type' should be f{ReactiveType.SUBMIT_MODAL}")
        return value


class RequestModalReactiveBody(BaseReactiveBody):
    react_user_id: int

    def __init__(self, **data) -> None:
        if 'type' not in data:
            data['type'] = ReactiveType.REQUEST_MODAL
        super().__init__(**data)

    @validator('type')
    def _check_type(cls, value: ReactiveType) -> ReactiveType:
        if value != ReactiveType.REQUEST_MODAL:
            raise ValueError(f"The 'type' should be f{ReactiveType.REQUEST_MODAL}")
        return value


class ModalReactiveView(BaseModel):
    title: str
    accept: str
    decline: str
    blocks: List[Block]
    value: str

    def __init__(self, *, blocks: Optional[List[Union[Block, Dict]]] = None, **data) -> None:
        if blocks is not None:
            data['blocks'] = [Block.new(block) for block in blocks]
        super().__init__(**data)


class RequestModalReactiveResponse(BaseModel):
    view: ModalReactiveView

    def __str__(self):
        return self.json(exclude_none=True)

    def __repr__(self):
        return str(self)

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, RequestModalReactiveResponse):
            return False
        return self.dict(exclude_none=True) == value.dict(exclude_none=True)


class BaseResponse(BaseModel, ABC):
    success: bool = True
    error: Optional[ErrorField] = None

    class Config:
        validate_assignment = True
        json_encoders = {
            datetime: lambda dt: int(dt.timestamp()),
        }

    def __str__(self):
        return self.json(exclude_none=True)

    def __repr__(self):
        return str(self)

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, BaseResponse):
            return False
        return self.dict(exclude_none=True) == value.dict(exclude_none=True)

    def plain(self) -> str:
        if self.error:
            return '\n'.join([
                f'Error code:\t{self.error.code}',
                f'Message:\t{self.error.message}',
            ])
        return 'OK'


class UserResponse(BaseResponse):
    user: Optional[UserField] = None

    def plain(self) -> str:
        if self.user:
            return '\n'.join([
                f'ID:\t{self.user.id}',
                f'Name:\t{self.user.name}',
                f'Nickname:\t{self.user.nickname}',
                f'Department:\t{self.user.department}',
                f'Position:\t{self.user.position}',
                f'Responsibility:\t{self.user.responsibility}',
                f'Tels:\t{self.user.tels}',
                f'Mobiles:\t{self.user.mobiles}',
                f'Work start time:\t{self.user.work_start_time}',
                f'Work end time:\t{self.user.work_end_time}',
                f'Vacation start time:\t{self.user.vacation_start_time}',
                f'Vacation end time:\t{self.user.vacation_end_time}',
            ])
        return super().plain()


class UserListResponse(BaseResponse):
    cursor: Optional[str] = None
    users: Optional[List[UserField]] = None

    def plain(self) -> str:
        if self.users:
            outs = []
            for user in self.users:
                out = '\n'.join([
                    f'ID:\t{user.id}',
                    f'Name:\t{user.name}',
                    f'Nickname:\t{user.nickname}',
                    f'Department:\t{user.department}',
                    f'Position:\t{user.position}',
                    f'Responsibility:\t{user.responsibility}',
                    f'Tels:\t{user.tels}',
                    f'Mobiles:\t{user.mobiles}',
                    f'Work start time:\t{user.work_start_time}',
                    f'Work end time:\t{user.work_end_time}',
                    f'Vacation start time:\t{user.vacation_start_time}',
                    f'Vacation end time:\t{user.vacation_end_time}',
                ])
                outs.append(out)
            return ('\n' + '-' * 30 + '\n').join(outs)
        return super().plain()


class ConversationResponse(BaseResponse):
    conversation: Optional[ConversationField] = None

    def plain(self) -> str:
        if self.conversation:
            return '\n'.join([
                f'ID:\t{self.conversation.id}',
                f'Type:\t{self.conversation.type}',
                f'Name:\t{self.conversation.name}',
                f'Avatar URL:\t{self.conversation.avatar_url}',
            ])
        return super().plain()


class ConversationListResponse(BaseResponse):
    cursor: Optional[str] = None
    conversations: Optional[List[ConversationField]] = None

    def plain(self) -> str:
        if self.conversations:
            outs = []
            for conversation in self.conversations:
                out = '\n'.join([
                    f'ID:\t{conversation.id}',
                    f'Type:\t{conversation.type}',
                    f'Name:\t{conversation.name}',
                    f'Avatar URL:\t{conversation.avatar_url}',
                ])
                outs.append(out)
            return ('\n' + '-' * 30 + '\n').join(outs)
        return super().plain()


class MessageResponse(BaseResponse):
    message: Optional[MessageField] = None

    def plain(self):
        if self.message:
            return '\n'.join([
                f'ID:\t{self.message.id}',
                f'Conversation ID:\t{self.message.conversation_id}',
                f'Send time:\t{self.message.send_time}',
                f'Update time:\t{self.message.update_time}',
                f'Text:\t{self.message.text}',
                f'Blocks:\t{len(self.message.blocks) if self.message.blocks else "-"}',
            ])
        return super().plain()


class DepartmentListResponse(BaseResponse):
    cursor: Optional[str] = None
    departments: Optional[List[DepartmentField]] = None

    def plain(self) -> str:
        if self.departments:
            outs = []
            for department in self.departments:
                out = '\n'.join([
                    f'ID:\t\t{department.id}',
                    f'Name:\t\t{department.name}',
                    f'Code:\t\t{department.code}',
                    f'User count:\t{department.user_count}',
                ])
                outs.append(out)
            return ('\n' + '-' * 30 + '\n').join(outs)
        return super().plain()


class SpaceResponse(BaseResponse):
    space: Optional[SpaceField] = None

    def plain(self) -> str:
        if self.space:
            return '\n'.join([
                f'ID:\t{self.space.id}',
                f'OrgID:\t{self.space.kakaoi_org_id}',
                f'Name:\t{self.space.name}',
                f'Color code:\t{self.space.color_code}',
                f'Color tone:\t{self.space.color_tone}',
                f'Permitted ext:\t{self.space.permitted_ext}',
                f'Profile name format:\t{self.space.profile_name_format}',
                f'Profile position format:\t{self.space.profile_position_format}',
                f'Logo URL:\t{self.space.logo_url}',
            ])
        return super().plain()


class BotResponse(BaseResponse):
    info: Optional[BotField] = None

    def plain(self) -> str:
        if self.info:
            return '\n'.join([f'ID:\t{self.info.bot_id}', f'Name:\t{self.info.title}', f'Status:\t{self.info.status}'])
        return super().plain()



import json
from enum import unique
from functools import reduce
from abc import ABC
from typing import Any, Dict, List, Optional, Type, Union, ClassVar, Iterable

from pydantic import (
    BaseModel,
    root_validator,
    validator,
    AnyHttpUrl,
)

from kakaowork.consts import StrEnum
from kakaowork.exceptions import InvalidBlock, InvalidBlockType


@unique
class BlockType(StrEnum):
    TEXT = "text"
    IMAGE_LINK = "image_link"
    BUTTON = "button"
    DIVIDER = "divider"
    HEADER = "header"
    ACTION = "action"
    DESCRIPTION = "description"
    SECTION = "section"
    CONTEXT = "context"
    LABEL = "label"
    INPUT = "input"
    SELECT = "select"

    @classmethod
    def block_cls(cls, type: Union[str, 'BlockType']) -> Type['Block']:
        bt = cls(type) if isinstance(type, str) else type
        if bt == cls.TEXT:
            return TextBlock
        elif bt == cls.IMAGE_LINK:
            return ImageLinkBlock
        elif bt == cls.BUTTON:
            return ButtonBlock
        elif bt == cls.DIVIDER:
            return DividerBlock
        elif bt == cls.HEADER:
            return HeaderBlock
        elif bt == cls.ACTION:
            return ActionBlock
        elif bt == cls.DESCRIPTION:
            return DescriptionBlock
        elif bt == cls.SECTION:
            return SectionBlock
        elif bt == cls.CONTEXT:
            return ContextBlock
        elif bt == cls.LABEL:
            return LabelBlock
        elif bt == cls.INPUT:
            return InputBlock
        elif bt == cls.SELECT:
            return SelectBlock
        raise InvalidBlockType()


class TextInlineType(StrEnum):
    STYLED = 'styled'
    LINK = 'link'


class TextInlineColor(StrEnum):
    DEFAULT = 'default'
    GREY = 'grey'
    BLUE = 'blue'
    RED = 'red'

    @classmethod
    def _missing_(cls, value: Any) -> 'TextInlineColor':
        return cls.DEFAULT


class TextInline(BaseModel):
    type: TextInlineType
    text: str
    bold: Optional[bool] = None
    italic: Optional[bool] = None
    strike: Optional[bool] = None
    color: Optional[TextInlineColor] = None
    url: Optional[str] = None

    class Config:
        underscore_attrs_are_private = True
        validate_assignment = True

    @root_validator
    def _check_conflict_props(cls, values: Dict) -> Dict:
        _type = values.get('type')
        if _type is TextInlineType.STYLED and values.get('url') is not None:
            raise ValueError("If the 'type' property is 'styled', the 'url' property can't be set")
        elif _type is TextInlineType.LINK:
            only_url_set = all([
                values.get('bold') is None,
                values.get('italic') is None,
                values.get('strike') is None,
                values.get('color') is None,
            ])
            if not only_url_set:
                raise ValueError("If the 'type' property is 'link', the 'url' property only can be set")
        return values


@unique
class ButtonStyle(StrEnum):
    DEFAULT = "default"
    PRIMARY = "primary"
    DANGER = "danger"

    @classmethod
    def _missing_(cls, value: Any) -> 'ButtonStyle':
        return cls.DEFAULT


@unique
class ButtonActionType(StrEnum):
    OPEN_INAPP_BROWSER = "open_inapp_browser"
    OPEN_SYSTEM_BROWSER = "open_system_browser"
    OPEN_EXTERNAL_APP = "open_external_app"
    SUBMIT_ACTION = "submit_action"
    CALL_MODAL = "call_modal"


@unique
class HeaderStyle(StrEnum):
    BLUE = 'blue'
    RED = 'red'
    YELLOW = 'yellow'

    @classmethod
    def _missing_(cls, value: Any) -> 'HeaderStyle':
        return cls.BLUE


@unique
class BlockKitType(StrEnum):
    MESSAGE = 'message'
    MODAL = 'modal'


class SelectBlockOption(BaseModel):
    text: str
    value: str


class Block(BaseModel, ABC):
    type: BlockType

    class Config:
        underscore_attrs_are_private = True
        validate_assignment = True

    def __str__(self):
        return self.json(exclude_none=True)

    def __repr__(self):
        return str(self)

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, Block):
            return False
        if self.type != value.type:
            return False
        if str(self) != str(value):
            return False
        return True

    @classmethod
    def new(cls, value: Union[Dict, 'Block']) -> 'Block':
        if isinstance(value, dict):
            if 'type' not in value:
                raise ValueError("There is no 'type' from the argument")
            return BlockType.block_cls(value['type'])(**value)
        return value


class TextBlock(Block):
    _max_len_text: ClassVar[int] = 500

    text: str
    markdown: Optional[bool] = None
    inlines: Optional[List[TextInline]] = None

    def __init__(self, **data) -> None:
        if 'type' not in data:
            data['type'] = BlockType.TEXT
        super().__init__(**data)

    @root_validator
    def _check_deprecation(cls, values: Dict) -> Dict:
        markdown = values.get('markdown')
        inlines = values.get('inlines')
        if markdown and inlines is not None:
            raise ValueError("The 'markdown' property can't be set with the 'inlines' property")
        return values

    @validator('text')
    def _check_max_len_text(cls, value: str) -> str:
        if not value:
            raise ValueError("The 'text' property should be exists")
        if len(value) > cls._max_len_text:
            raise ValueError(f"The 'text' property's length should be less than or equal to {cls._max_len_text}")
        return value

    @validator('inlines')
    def _check_max_len_inlines(cls, value: Optional[List[TextInline]]) -> Optional[List[TextInline]]:
        if value is not None:
            len_inlines = reduce(lambda acc, text: acc + len(text), map(lambda i: i.text, value), 0)
            if len_inlines > cls._max_len_text:
                raise ValueError(f"The 'inlines' property's all texts should be less than or equal to {cls._max_len_text}")
        return value


class ImageLinkBlock(Block):
    url: AnyHttpUrl

    def __init__(self, **data) -> None:
        if 'type' not in data:
            data['type'] = BlockType.IMAGE_LINK
        super().__init__(**data)


class ButtonBlock(Block):
    _max_len_text: ClassVar[int] = 20

    text: str
    style: ButtonStyle = ButtonStyle.DEFAULT
    action_type: Optional[ButtonActionType] = None
    action_name: Optional[str] = None
    value: Optional[str] = None

    def __init__(self, **data) -> None:
        if 'type' not in data:
            data['type'] = BlockType.BUTTON
        super().__init__(**data)

    @validator('text')
    def _check_text(cls, value: str) -> str:
        if not value:
            raise ValueError("The 'text' property should be exists")
        if len(value) > cls._max_len_text:
            raise ValueError(f"The 'text' property's length should be less than or equal to {cls._max_len_text}")
        return value


class DividerBlock(Block):
    def __init__(self, **data) -> None:
        if 'type' not in data:
            data['type'] = BlockType.DIVIDER
        super().__init__(**data)


class HeaderBlock(Block):
    _max_len_text: ClassVar[int] = 20

    text: str
    style: HeaderStyle = HeaderStyle.BLUE

    def __init__(self, **data):
        if 'type' not in data:
            data['type'] = BlockType.HEADER
        super().__init__(**data)

    @validator('text')
    def _check_text(cls, value: str) -> str:
        if not value:
            raise ValueError("The 'text' property should be exists")
        if len(value) > cls._max_len_text:
            raise ValueError(f"The 'text' property's length should be less than or equal to {cls._max_len_text}")
        return value


class ActionBlock(Block):
    _max_len_elements: ClassVar[int] = 3

    elements: List[ButtonBlock]

    def __init__(self, **data) -> None:
        if 'type' not in data:
            data['type'] = BlockType.ACTION
        super().__init__(**data)

    @validator('elements')
    def _check_elements(cls, value: List[ButtonBlock]) -> List[ButtonBlock]:
        if not value:
            raise ValueError("The 'elements' property should be exists")
        if len(value) > cls._max_len_elements:
            raise ValueError(f"The 'elements' property's length should be less than or equal to {cls._max_len_elements}")
        return value


class DescriptionBlock(Block):
    _max_len_term: ClassVar[int] = 10

    term: str
    content: TextBlock
    accent: Optional[bool] = None

    def __init__(self, **data) -> None:
        if 'type' not in data:
            data['type'] = BlockType.DESCRIPTION
        super().__init__(**data)

    @validator('term')
    def _check_term(cls, value: str) -> str:
        if not value:
            raise ValueError("The 'term' property should be exists")
        if len(value) > cls._max_len_term:
            raise ValueError(f"The 'term' property's length should be less than or equal to {cls._max_len_term}")
        return value


class SectionBlock(Block):
    content: TextBlock
    accessory: ImageLinkBlock

    def __init__(self, **data) -> None:
        if 'type' not in data:
            data['type'] = BlockType.SECTION
        super().__init__(**data)


class ContextBlock(Block):
    content: TextBlock
    image: ImageLinkBlock

    def __init__(self, **data) -> None:
        if 'type' not in data:
            data['type'] = BlockType.CONTEXT
        super().__init__(**data)


class LabelBlock(Block):
    _max_len_text: ClassVar[int] = 200

    text: str
    markdown: bool

    def __init__(self, **data) -> None:
        if 'type' not in data:
            data['type'] = BlockType.LABEL
        super().__init__(**data)

    @validator('text')
    def _check_text(cls, value: str) -> str:
        if not value:
            raise ValueError("The 'text' property should be exists")
        if len(value) > cls._max_len_text:
            raise ValueError(f"The 'text' property's length should be less than or equal to {cls._max_len_text}")
        return value


class InputBlock(Block):
    _max_len_placeholder: ClassVar[int] = 50

    name: str
    required: Optional[bool] = None
    placeholder: Optional[str] = None

    def __init__(self, **data) -> None:
        if 'type' not in data:
            data['type'] = BlockType.INPUT
        super().__init__(**data)

    @validator('name')
    def _check_name(cls, value: str) -> str:
        if not value:
            raise ValueError("The 'name' property should be exists")
        return value

    @validator('placeholder')
    def _check_placeholder(cls, value: Optional[str] = None) -> Optional[str]:
        if value and len(value) > cls._max_len_placeholder:
            raise ValueError(f"The 'placeholder' property's length should be less than or equal to {cls._max_len_placeholder}")
        return value


class SelectBlock(Block):
    _max_len_options: ClassVar[int] = 30
    _max_len_placeholder: ClassVar[int] = 50

    name: str
    options: List[SelectBlockOption]
    required: Optional[bool] = None
    placeholder: Optional[str] = None

    def __init__(self, **data) -> None:
        if 'type' not in data:
            data['type'] = BlockType.SELECT
        super().__init__(**data)

    @validator('name')
    def _check_name(cls, value: str) -> str:
        if not value:
            raise ValueError("The 'name' property should be exists")
        return value

    @validator('options')
    def _check_options(cls, value: List[SelectBlockOption]) -> List[SelectBlockOption]:
        if not value:
            raise ValueError("The 'options' property should be exists")
        if len(value) > cls._max_len_options:
            raise ValueError(f"The 'options' property's length should be less than or equal to {cls._max_len_options}")
        return value

    @validator('placeholder')
    def _check_placeholder(cls, value: Optional[str] = None) -> Optional[str]:
        if value and len(value) > cls._max_len_placeholder:
            raise ValueError(f"The 'placeholder' property's length should be less than or equal to {cls._max_len_placeholder}")
        return value


class BlockKitBuilder(BaseModel):
    type: BlockKitType
    blocks: List[Block] = []
    text: Optional[str] = None
    title: Optional[str] = None
    accept: Optional[str] = None
    decline: Optional[str] = None
    value: Optional[str] = None

    class Config:
        validate_assignment = True

    @root_validator
    def _check_message_type(cls, values: Dict) -> Dict:
        _type = values.get('type')
        if _type and _type != BlockKitType.MESSAGE:
            if values.get('text'):
                raise ValueError("The 'text' property can be set only for message type")
        return values

    @root_validator
    def _check_modal_type(cls, values: Dict) -> Dict:
        _type = values.get('type')
        if _type and _type != BlockKitType.MODAL:
            if values.get('title'):
                raise ValueError("The 'title' property can be set only for modal type")
            elif values.get('accept'):
                raise ValueError("The 'accept' property can be set only for modal type")
            elif values.get('decline'):
                raise ValueError("The 'decline' property can be set only for modal type")
            elif values.get('value'):
                raise ValueError("The 'value' property can be set only for modal type")
        return values

    def add_block(self, block: Union[Block, dict]):
        if isinstance(block, dict):
            if 'type' not in block:
                raise InvalidBlock()
            block = BlockType.block_cls(block['type'])(**block)
        self.blocks.append(block)

    @classmethod
    def load(cls, path: str) -> 'BlockKitBuilder':
        with open(path, 'r') as f:
            data = json.load(f)
        self = cls(type=data['type'])
        for key, value in data.items():
            if key == 'blocks' and isinstance(value, Iterable):
                self.blocks = []
                for block in value:
                    self.add_block(block)
            else:
                setattr(self, key, value)
        return self



# noqa: D104
from kakaowork.blockkit import (
    Block,
    BlockType,
    ButtonStyle,
    ButtonActionType,
    HeaderStyle,
    BlockKitType,
    SelectBlockOption,
    TextInlineType,
    TextInlineColor,
    TextInline,
    TextBlock,
    ImageLinkBlock,
    ButtonBlock,
    DividerBlock,
    HeaderBlock,
    ActionBlock,
    DescriptionBlock,
    SectionBlock,
    ContextBlock,
    LabelBlock,
    InputBlock,
    SelectBlock,
    BlockKitBuilder,
)

from kakaowork.consts import (
    Limit,
    BASE_URL,
    BASE_PATH_USERS,
    BASE_PATH_CONVERSATIONS,
    BASE_PATH_MESSAGES,
    BASE_PATH_DEPARTMENTS,
    BASE_PATH_SPACES,
    BASE_PATH_BOTS,
    BASE_PATH_BATCH,
)

from kakaowork.exceptions import (KakaoworkError, InvalidBlock, InvalidBlockType)

from kakaowork.client import (Kakaowork, AsyncKakaowork)

from kakaowork.models import (
    ErrorCode,
    ConversationType,
    ColorTone,
    ProfileNameFormat,
    ProfilePositionFormat,
    BotStatus,
    ErrorField,
    UserIdentificationField,
    UserField,
    ConversationField,
    MessageField,
    DepartmentField,
    SpaceField,
    BotField,
    WorkTimeField,
    VacationTimeField,
    BaseResponse,
    UserResponse,
    UserListResponse,
    ConversationResponse,
    ConversationListResponse,
    MessageResponse,
    DepartmentListResponse,
    SpaceResponse,
    BotResponse,
    ReactiveType,
    BaseReactiveBody,
    SubmitActionReactiveBody,
    SubmitModalReactiveBody,
    RequestModalReactiveBody,
    ModalReactiveView,
    RequestModalReactiveResponse,
)

from kakaowork.reactive import (
    BaseReactiveActionHandler,
    BaseReactiveModalHandler,
)

__version__ = '0.7.0'



import time
import asyncio
from threading import RLock
from typing import Optional, Callable, List, Dict
from functools import wraps


class RateLimiter:
    """Rate limiter class."""
    def __init__(self, *, capacity: int, refill_rate: float) -> None:
        """Initialize the rate limiter.

        Args:
            capacity: Maximum number of requests per second.
            refill_rate: Rate at which the rate limiter refills its capacity.
        """
        self._capacity = capacity
        self._refill_rate = refill_rate
        self._timer = time.perf_counter  # Ref https://www.webucator.com/article/python-clocks-explained/
        self._lock = RLock()
        self._last_refill_time = self._timer()
        self._tokens = capacity

    def __call__(self, f: Callable):
        """Decorator to rate limit a function.

        Args:
            f: Function to rate limit.
        """
        @wraps(f)
        def wrapper(*args: List, **kwargs: Dict):
            with self:
                return f(*args, **kwargs)

        return wrapper

    def __enter__(self) -> 'RateLimiter':
        """Enter the rate limiter."""
        if self.capacity <= 0:
            return self
        wait_time = self.limit()
        if wait_time > 0.0:
            time.sleep(wait_time)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the rate limiter."""
        pass

    async def __aenter__(self) -> 'RateLimiter':
        """Enter the rate limiter."""
        if self.capacity <= 0:
            return self
        wait_time = self.limit()
        if wait_time > 0.0:
            await asyncio.sleep(wait_time)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Exit the rate limiter."""
        pass

    @property
    def capacity(self) -> int:  # noqa: D102
        with self._lock:
            return self._capacity

    @capacity.setter
    def capacity(self, value: int) -> None:  # noqa: D102
        with self._lock:
            self._capacity = value

    @capacity.deleter
    def capacity(self) -> None:  # noqa: D102
        with self._lock:
            self._capacity = 0

    @property
    def refill_rate(self) -> float:  # noqa: D102
        with self._lock:
            return self._refill_rate

    @refill_rate.setter
    def refill_rate(self, value: float) -> None:  # noqa: D102
        with self._lock:
            self._refill_rate = value

    @refill_rate.deleter
    def refill_rate(self) -> None:  # noqa: D102
        with self._lock:
            self._refill_rate = 0.0

    def reset(self, *, capacity: Optional[int] = None, refill_rate: Optional[float] = None) -> None:
        """Reset the rate limiter.

        Args:
            capacity: A capacity after limiter reset
            refill_rate: A refill rate after limiter reset
        """
        with self._lock:
            if capacity is not None:
                self._capacity = capacity
            if refill_rate is not None:
                self._refill_rate = refill_rate
            self._last_refill_time = self._timer()
            self._tokens = self._capacity

    def limit(self, requests: int = 1) -> float:
        """Limit the rate of requests.

        Args:
            requests: Number of requests

        Returns:
            A float number of seconds to wait before next request
        """
        with self._lock:
            refill_time = self._timer()
            refill_tokens = int((refill_time - self._last_refill_time) / self.refill_rate)
            if refill_tokens > 0:
                self._tokens = min(self._capacity, self._tokens + refill_tokens)
                self._last_refill_time = refill_time

            wait_time = 0.0
            if self._tokens - requests >= 0:
                self._tokens -= requests
            else:
                wait_time = abs(self._tokens - requests) * self.refill_rate
            return wait_time
